from typing import Type, Dict

from dagshub.data_engine.dtypes import MetadataFieldType

metadataTypeLookup = {
    int: MetadataFieldType.INTEGER,
    bool: MetadataFieldType.BOOLEAN,
    float: MetadataFieldType.FLOAT,
    str: MetadataFieldType.STRING,
    bytes: MetadataFieldType.BLOB,
}

metadataTypeLookupReverse: Dict[str, Type] = {}
for k, v in metadataTypeLookup.items():
    metadataTypeLookupReverse[v.value] = k



import inspect
import logging
import multiprocessing.pool
from dataclasses import field, dataclass
from itertools import repeat
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, List, Dict, Any, Optional, Union

import rich.progress

from dagshub.common import config
from dagshub.common.analytics import send_analytics_event
from dagshub.common.download import download_files
from dagshub.common.helpers import sizeof_fmt, prompt_user
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import lazy_load
from dagshub.data_engine.annotation.voxel_conversion import (
    add_voxel_annotations,
    add_ls_annotations,
)
from dagshub.data_engine.client.models import DatasourceType
from dagshub.data_engine.model.datapoint import Datapoint, _get_blob
from dagshub.data_engine.client.loaders.base import DagsHubDataset
from dagshub.data_engine.voxel_plugin_server.utils import set_voxel_envvars
from dagshub.data_engine.dtypes import MetadataFieldType

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource
    import fiftyone as fo
    import dagshub.data_engine.voxel_plugin_server.server as plugin_server_module
else:
    plugin_server_module = lazy_load("dagshub.data_engine.voxel_plugin_server.server")
    fo = lazy_load("fiftyone")
    tf = lazy_load("tensorflow")

logger = logging.getLogger(__name__)


class VisualizeError(Exception):
    """:meta private:"""

    pass


@dataclass
class QueryResult:
    """
    Result of executing a query on a :class:`.Datasource`.

    You can iterate over this object to get the :class:`datapoints <.Datapoint>`::

        res = ds.head()
        for dp in res:
            print(dp.path_in_repo)

    """

    _entries: List[Datapoint]
    datasource: "Datasource"
    _datapoint_path_lookup: Dict[str, Datapoint] = field(init=False)

    def __post_init__(self):
        self._refresh_lookups()

    @property
    def entries(self):
        """
        list(Datapoint): Datapoints contained in this QueryResult
        """
        return self._entries

    @entries.setter
    def entries(self, value: List[Datapoint]):
        self._entries = value
        self._refresh_lookups()

    def _refresh_lookups(self):
        self._datapoint_path_lookup = {}
        for e in self.entries:
            self._datapoint_path_lookup[e.path] = e

    @property
    def dataframe(self):
        """
        Represent the contents of this QueryResult as a
        `pandas.DataFrame <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_.

        The created dataframe has a copy of the QueryResult's data.
        """
        import pandas as pd

        metadata_keys = set()
        for e in self.entries:
            metadata_keys.update(e.metadata.keys())

        metadata_keys = list(sorted(metadata_keys))
        return pd.DataFrame.from_records([dp.to_dict(metadata_keys) for dp in self.entries])

    def __len__(self):
        return len(self.entries)

    def __iter__(self):
        """You can iterate over a QueryResult to get containing datapoints"""
        return self.entries.__iter__()

    def __repr__(self):
        return f"QueryResult of datasource {self.datasource.source.name} with {len(self.entries)} datapoint(s)"

    @staticmethod
    def from_gql_query(query_resp: Dict[str, Any], datasource: "Datasource") -> "QueryResult":
        if "edges" not in query_resp:
            return QueryResult([], datasource)
        if query_resp["edges"] is None:
            return QueryResult([], datasource)
        return QueryResult(
            [Datapoint.from_gql_edge(edge, datasource) for edge in query_resp["edges"]],
            datasource,
        )

    def as_ml_dataset(self, flavor: str, **kwargs):
        """
        Convert the QueryResult into a dataset for a machine learning framework

        Args:
            flavor: Either of:

                - ``"torch"``: returns a \
                `torch.utils.data.Dataset <https://pytorch.org/docs/stable/data.html#torch.utils.data.Dataset>`_
                - ``"tensorflow"``: returns a \
                `tf.data.Dataset <https://www.tensorflow.org/api_docs/python/tf/data/Dataset>`_


        Keyword Args:
            metadata_columns (list(str)): which fields to use in the dataset.
            strategy (str): Datapoint file loading strategy. Possible values:

                - ``"preload"`` - Download all datapoints before returning the dataset.
                - ``"background"`` - Start downloading the datapoints in the background. \
                    If an undownloaded datapoint is accessed, it gets downloaded.
                - ``"lazy"`` (default) - Download each datapoint as it is being accessed by the dataset.

            savedir (str|Path): Where to store the datapoint files. Default is :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`
            processes (int): number of parallel processes to download the datapoints with. Default is 8.
            tensorizers: How to transform the datapoint file/metadata into tensors. Possible values:

                - ``"auto"`` - try to guess the tensorizers for every field.\
                    For files the tensorizer will be the determined by the first file's extension.
                - ``"image"`` | ``"audio"`` - tensorize all fields according to this type
                - List of tensorizers. First is the tensorizer for the datapoint file and then \
                    a tensorizer for each of the metadata fields. Tensorizer can either be strings\
                     ``"image"``, ``"audio"``, ``"numeric"``, or your own function that receives the\
                     metadata value of the field and turns it into a tensor.
        """
        flavor = flavor.lower()
        if flavor == "torch":
            from dagshub.data_engine.client.loaders.torch import PyTorchDataset

            return PyTorchDataset(self, **kwargs)
        elif flavor == "tensorflow":
            from dagshub.data_engine.client.loaders.tf import TensorFlowDataset

            ds_builder = TensorFlowDataset(self, **kwargs)
            ds = tf.data.Dataset.from_generator(ds_builder.generator, output_signature=ds_builder.signature)
            ds.__len__ = lambda: ds_builder.__len__()
            ds.__getitem__ = ds_builder.__getitem__
            ds.builder = ds_builder
            ds.type = "tensorflow"
            return ds
        else:
            raise ValueError("supported flavors are torch|tensorflow")

    def as_ml_dataloader(self, flavor, **kwargs):
        """
        Convert the QueryResult into a dataloader for a machine learning framework

        Args:
            flavor: Either of:

                - ``"torch"``: returns a \
                `torch.utils.data.DataLoader <https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader>`_
                - ``"tensorflow"``: returns a \
                `tf.keras.utils.Sequence <https://www.tensorflow.org/api_docs/python/tf/keras/utils/Sequence>`_

        Kwargs are the same as :func:`as_ml_dataset()`.
        """

        send_analytics_event(
            "Client_DataEngine_DataLoaderInitialized",
            repo=self.datasource.source.repoApi,
        )

        def keypairs(keys):
            return {key: kwargs[key] for key in keys}

        if type(flavor) is not str:
            if flavor.type == "torch":
                from dagshub.data_engine.client.loaders.torch import PyTorchDataLoader

                return PyTorchDataLoader(flavor, **kwargs)
            elif flavor.type == "tensorflow":
                from dagshub.data_engine.client.loaders.tf import TensorFlowDataLoader

                return TensorFlowDataLoader(flavor, **kwargs)

        kwargs["for_dataloader"] = True
        dataset_kwargs = set(list(inspect.signature(DagsHubDataset).parameters.keys())[1:])
        global_kwargs = set(kwargs.keys())
        flavor = flavor.lower() if type(flavor) is str else flavor
        if flavor == "torch":
            from dagshub.data_engine.client.loaders.torch import PyTorchDataLoader

            return PyTorchDataLoader(
                self.as_ml_dataset(flavor, **keypairs(global_kwargs.intersection(dataset_kwargs))),
                **keypairs(global_kwargs - dataset_kwargs),
            )
        elif flavor == "tensorflow":
            from dagshub.data_engine.client.loaders.tf import TensorFlowDataLoader

            return TensorFlowDataLoader(
                self.as_ml_dataset(
                    flavor,
                    **keypairs(global_kwargs.intersection(dataset_kwargs)),
                ),
                **keypairs(global_kwargs - dataset_kwargs),
            )
        else:
            raise ValueError("supported flavors are torch|tensorflow|<torch.utils.data.Dataset>|<tf.data.Dataset>")

    def __getitem__(self, item: Union[str, int, slice]):
        """
        Gets datapoint by its path (string) or by its index in the result (or slice)
        """
        if type(item) is str:
            return self._datapoint_path_lookup[item]
        elif type(item) is int:
            return self.entries[item]
        elif type(item) is slice:
            return QueryResult(_entries=self.entries[item], datasource=self.datasource)
        else:
            raise ValueError(
                f"Can't lookup datapoint using value {item} of type {type(item)}, "
                f"needs to be either int or str or slice"
            )

    def get_blob_fields(
        self, *fields: str, load_into_memory=False, cache_on_disk=True, num_proc: int = config.download_threads
    ) -> "QueryResult":
        """
        Downloads data from blob fields

        Args:
            fields: list of binary fields to download blobs for. If empty, download all blob fields.
            load_into_memory: Whether to load the blobs into the datapoints, or just store them on disk


                If True: the datapoints' specified fields will contain the blob data

                If False: the datapoints' specified fields will contain Path objects to the file of the downloaded blob
            cache_on_disk: Whether to cache the blobs on disk or not (valid only if load_into_memory is set to True)
                Cache location is ``~/dagshub/datasets/<repo>/<datasource_id>/.metadata_blobs/``
            num_proc: number of download threads
        """
        send_analytics_event("Client_DataEngine_downloadBlobs", repo=self.datasource.source.repoApi)
        if not load_into_memory:
            assert cache_on_disk

        # If no fields are specified, include all blob fields from self.datasource.fields
        if not fields:
            fields = [field.name for field in self.datasource.fields if field.valueType == MetadataFieldType.BLOB]

        for fld in fields:
            logger.info(f"Downloading metadata for field {fld} with {num_proc} processes")
            cache_location = self.datasource.default_dataset_location / ".metadata_blobs"

            if cache_on_disk:
                cache_location.mkdir(parents=True, exist_ok=True)

            blob_urls = list(map(lambda dp: dp._extract_blob_url_and_path(fld), self.entries))
            auth = self.datasource.source.repoApi.auth
            func_args = zip(
                map(lambda bu: bu[0], blob_urls),
                map(lambda bu: bu[1], blob_urls),
                repeat(auth),
                repeat(cache_on_disk),
                repeat(load_into_memory),
            )
            with multiprocessing.pool.ThreadPool(num_proc) as pool:
                res = pool.starmap(_get_blob, func_args)

            for dp, binary_val_or_path in zip(self.entries, res):
                if binary_val_or_path is None:
                    continue
                dp.metadata[fld] = binary_val_or_path

        return self

    def download_binary_columns(
        self, *columns: str, load_into_memory=True, cache_on_disk=True, num_proc: int = 32
    ) -> "QueryResult":
        """
        deprecated: Use get_blob_fields instead.

        :meta private:
        """
        return self.get_blob_fields(
            *columns, load_into_memory=load_into_memory, cache_on_disk=cache_on_disk, num_proc=num_proc
        )

    def download_files(
        self,
        target_dir: Optional[Union[str, PathLike]] = None,
        keep_source_prefix=True,
        redownload=False,
        path_field: Optional[str] = None,
    ) -> PathLike:
        """
        Downloads the datapoints to the ``target_dir`` directory

        Args:
            target_dir: Where to download the files. Defaults to \
                :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`
            keep_source_prefix: If True, includes the prefix of the datasource in the download path.
            redownload: Whether to redownload a file if it exists on the filesystem already.\
                We don't do any hashsum checks, so if it's possible that the file has been updated, set to True
            path_field: Set this to the name of the field with the file's path\
                if you want to download files from a field other than the datapoint's path.

        .. note::
            For ``path_field`` the path in the field still needs to be in the same repo
            and have the same format as the path of the datapoint, including not having the prefix.
            For now, you can't download arbitrary paths/urls.

        Returns:
            Path to the directory with the downloaded files
        """

        target_path = self.datasource.default_dataset_location if target_dir is None else Path(target_dir).expanduser()
        logger.warning(f"Downloading {len(self.entries)} files to {str(target_path)}")

        if self.datasource.source.source_type == DatasourceType.BUCKET:
            send_analytics_event(
                "Client_DataEngine_downloadDatapointsFromBucket",
                repo=self.datasource.source.repoApi,
            )
        else:
            send_analytics_event(
                "Client_DataEngine_downloadDatapoints",
                repo=self.datasource.source.repoApi,
            )

        def dp_path(dp: Datapoint):
            if path_field is not None:
                path_val = dp.metadata.get(path_field)
                if path_val is None:
                    return None
            else:
                path_val = dp.path

            if keep_source_prefix:
                return target_path / self.datasource.source.source_prefix / path_val
            else:
                return target_path / path_val

        def dp_url(dp: Datapoint):
            if path_field is not None:
                path_val = dp.metadata.get(path_field)
                if path_val is None:
                    return None
                return self.datasource.source.raw_path(path_val)
            else:
                return dp.download_url

        download_args = [(dp_url(dp), dp_path(dp)) for dp in self.entries if dp_path(dp) is not None]

        download_files(download_args, skip_if_exists=not redownload)
        return target_path

    def to_voxel51_dataset(self, **kwargs) -> "fo.Dataset":
        """
        Creates a voxel51 dataset that can be used with\
        `fo.launch_app()
        <https://docs.voxel51.com/api/fiftyone.core.session.html?highlight=launch_app#fiftyone.core.session.launch_app>`_
        to visualize it.

        Keyword Args:
            name (str): Name of the dataset. Default is the name of the datasource.
            force_download (bool): Download the dataset even if the size of the files is bigger than 100MB.\
                Default is False
            files_location (str|PathLike): path to the location where to download the local files.
                Default is :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`
            redownload (bool): Redownload files, replacing the ones that might exist on the filesystem.\
                Default is False.
            voxel_annotations (List[str]) : List of fields from which to load voxel annotations serialized with
                                        `to_json()`. This will override the labelstudio annotations

        :rtype: `fo.Dataset <https://docs.voxel51.com/api/fiftyone.core.dataset.html#fiftyone.core.dataset.Dataset>`_
        """
        if len(self.entries) == 0:
            raise VisualizeError("No datapoints to visualize")
        logger.info("Migrating dataset to voxel51")
        name = kwargs.get("name", self.datasource.source.name)
        force_download = kwargs.get("force_download", False)
        # TODO: don't override samples in existing one
        if fo.dataset_exists(name):
            ds: fo.Dataset = fo.load_dataset(name)
        else:
            ds: fo.Dataset = fo.Dataset(name)
        # ds.persistent = True

        dataset_location = Path(kwargs.get("files_location", self.datasource.default_dataset_location))
        dataset_location.mkdir(parents=True, exist_ok=True)
        logger.info("Downloading files...")

        if "voxel_annotations" in kwargs:
            annotation_fields = kwargs["voxel_annotations"]
            label_func = add_voxel_annotations
        else:
            annotation_fields = self.datasource.annotation_fields
            label_func = add_ls_annotations

        # Load the annotation fields
        self.download_binary_columns(*annotation_fields)

        if not force_download:
            self._check_downloaded_dataset_size()

        redownload = kwargs.get("redownload", False)
        self.download_files(self.datasource.default_dataset_location, redownload=redownload)

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        task = progress.add_task("Generating voxel samples...", total=len(self.entries))

        samples: List["fo.Sample"] = []

        with progress:
            for datapoint in self.entries:
                filepath = self.datasource.default_dataset_location / datapoint.path_in_repo
                sample = fo.Sample(filepath=filepath)
                sample["dagshub_download_url"] = datapoint.download_url
                sample["datapoint_id"] = datapoint.datapoint_id
                sample["datapoint_path"] = datapoint.path
                label_func(sample, datapoint, *annotation_fields)
                for k, v in datapoint.metadata.items():
                    # TODO: more filtering here, not all fields should be showing up in voxel
                    if k in annotation_fields:
                        continue
                    if type(v) is not bytes:
                        sample[k] = v
                samples.append(sample)
                progress.update(task, advance=1, refresh=True)

        ds.merge_samples(samples)
        return ds

    def _check_downloaded_dataset_size(self):
        download_size_prompt_threshold = 100 * (2**20)  # 100 Megabytes
        dp_size = self._calculate_datapoint_size()
        if dp_size is not None and dp_size > download_size_prompt_threshold:
            prompt = f"You're about to download {sizeof_fmt(dp_size)} of images locally."
            should_download = prompt_user(prompt)
            if not should_download:
                msg = "Downloading voxel dataset cancelled"
                logger.warning(msg)
                raise RuntimeError(msg)

    def _calculate_datapoint_size(self) -> Optional[int]:
        sum_size = 0
        has_sum_field = False
        all_have_sum_field = True
        size_field = "size"
        for dp in self.entries:
            if size_field in dp.metadata:
                has_sum_field = True
                sum_size += dp.metadata[size_field]
            else:
                all_have_sum_field = False
        if not has_sum_field:
            logger.warning("None of the datapoints had a size field, can't calculate size of the downloading dataset")
            return None
        if not all_have_sum_field:
            logger.warning("Not every datapoint has a size field, size calculations might be wrong")
        return sum_size

    def visualize(self, **kwargs):
        """
        Visualize this QueryResult with Voxel51.

        This function calls :func:`to_voxel51_dataset`, passing to it the keyword arguments,
        and launches a fiftyone session showing the dataset.

        Additionally, this function adds a DagsHub plugin into Voxel51 that you can use for additional interactions
        with the datasource from within the voxel environment.

        Returns the session object that you can ``wait()`` if you are using it\
        outside a notebook and need to not close the script immediately::

            session = ds.all().visualize()
            session.wait(-1)
        """
        set_voxel_envvars()

        send_analytics_event("Client_DataEngine_VizualizeResults", repo=self.datasource.source.repoApi)

        ds = self.to_voxel51_dataset(**kwargs)

        sess = fo.launch_app(ds)
        # Launch the server for plugin interaction
        plugin_server_module.run_plugin_server(sess, self.datasource, self.datasource.source.revision)

        return sess

    def annotate(
        self, open_project=True, ignore_warning=True, fields_to_embed=None, fields_to_exclude=None
    ) -> Optional[str]:
        """
        Sends all the datapoints returned in this QueryResult to be annotated in Label Studio on DagsHub.

        Args:
            open_project: Automatically open the Label Studio project in the browser
            ignore_warning: Suppress the prompt-warning if you try to annotate too many datapoints at once.
            fields_to_embed: list of meta-data columns that will show up in Label Studio UI.
             if not specified all will be displayed.
            fields_to_exclude: list of meta-data columns that will not show up in Label Studio UI
        Returns:
            The URL of the created Label Studio workspace
        """
        send_analytics_event("Client_DataEngine_SentToAnnotation", repo=self.datasource.source.repoApi)

        return self.datasource.send_datapoints_to_annotation(
            self.entries,
            open_project=open_project,
            ignore_warning=ignore_warning,
            fields_to_exclude=fields_to_exclude,
            fields_to_embed=fields_to_embed,
        )



import enum
from abc import ABCMeta
from typing import Set


class ReservedTags(enum.Enum):
    """:meta private:"""

    ANNOTATION = "annotation"


# These are the base primitives that the data engine database is capable of storing
class MetadataFieldType(enum.Enum):
    """
    Backing types in the Data Engine's database
    """

    BOOLEAN = "BOOLEAN"
    """Python's ``bool``"""
    INTEGER = "INTEGER"
    """Python's ``int``"""
    FLOAT = "FLOAT"
    """Python's ``float``"""
    STRING = "STRING"
    """Python's ``str``"""
    BLOB = "BLOB"
    """Python's ``bytes``"""


class DagshubDataType(metaclass=ABCMeta):
    """
    Inheritors of this ABC define custom types

    They are backed by a primitive type, but they also may have additional tags, which we use to enhance the experience.

    Attributes:
        backing_field_type: primitive type in the data engine database
        custom_tags: additional tags applied to this type
    """

    backing_field_type: MetadataFieldType = None
    custom_tags: Set[str] = None


class Int(DagshubDataType):
    """Basic python ``int``"""

    backing_field_type = MetadataFieldType.INTEGER


class String(DagshubDataType):
    """Basic python ``str``"""

    backing_field_type = MetadataFieldType.STRING


class Blob(DagshubDataType):
    """
    Basic python ``bytes``

    .. note::
        DagsHub doesn't return the blob fields by default, instead returning their hashes.
        Check out :func:`Datapoint.get_blob() <dagshub.data_engine.model.datapoint.Datapoint.get_blob>`
        to learn how to download the blob value.
    """

    backing_field_type = MetadataFieldType.BLOB


class Float(DagshubDataType):
    """
    Basic python ``float``
    """

    backing_field_type = MetadataFieldType.FLOAT


class Bool(DagshubDataType):
    """
    Basic python ``bool``
    """

    backing_field_type = MetadataFieldType.BOOLEAN


class LabelStudioAnnotation(DagshubDataType):
    """
    LabelStudio annotation. Backing type is blob.
    Has the annotation tag set.
    """

    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.ANNOTATION.value}


class Voxel51Annotation(DagshubDataType):
    """
    Voxel51 annotation. Backing type is blob.
    Has the annotation tag set.
    """

    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.ANNOTATION.value}



import logging
from os import PathLike
from pathlib import Path, PurePosixPath

import rich.progress

from dagshub.common.api.responses import (
    RepoAPIResponse,
    BranchAPIResponse,
    CommitAPIResponse,
    StorageAPIEntry,
    ContentAPIEntry,
    StorageContentAPIResult,
)
from dagshub.common.download import download_files
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import multi_urljoin
from functools import partial

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

from typing import Optional, Tuple, Any, List, Union

import dacite

import dagshub.auth
from dagshub.common import config

from dagshub.common.helpers import http_request, log_message

logger = logging.getLogger("dagshub")


class WrongRepoFormatError(Exception):
    pass


class RepoNotFoundError(Exception):
    pass


class BranchNotFoundError(Exception):
    pass


class CommitNotFoundError(Exception):
    pass


class PathNotFoundError(Exception):
    pass


class RepoAPI:
    def __init__(self, repo: str, host: Optional[str] = None, auth: Optional[Any] = None):
        """
        Class for interacting with the API of a repository

        Args:
            repo: repo in the format of ``user/repo``
            host (optional): url of the DagsHub instance the repo is on
            auth (optional): authentication to use to connect
        """
        self.owner, self.repo_name = self.parse_repo(repo)
        self.host = host if host is not None else config.host

        if auth is None:
            self.auth = dagshub.auth.get_authenticator(host=host)
        else:
            self.auth = auth

    def _http_request(self, method, url, **kwargs):
        if "auth" not in kwargs:
            kwargs["auth"] = self.auth
        return http_request(method, url, **kwargs)

    def get_repo_info(self) -> RepoAPIResponse:
        """
        Get information about the repository
        """
        res = self._http_request("GET", self.repo_api_url)

        if res.status_code == 404:
            raise RepoNotFoundError(f"Repo {self.repo_url} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting repository info."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return dacite.from_dict(RepoAPIResponse, res.json())

    def get_branch_info(self, branch: str) -> BranchAPIResponse:
        """
        Get information about specified branch

        Args:
            branch: Name of the branch to get the info
        """
        res = self._http_request("GET", self.branch_url(branch))

        if res.status_code == 404:
            raise BranchNotFoundError(f"Branch {branch} not found in repo {self.repo_url}")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting branch."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        return dacite.from_dict(BranchAPIResponse, res.json())

    def get_commit_info(self, sha: str) -> CommitAPIResponse:
        """
        Get information about a specific commit

        Args:
            sha: SHA of the commit.
        """
        res = self._http_request("GET", self.commit_url(sha))

        if res.status_code == 404:
            raise BranchNotFoundError(f"Commit {sha} not found in repo {self.repo_url}")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting commit."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        return dacite.from_dict(CommitAPIResponse, res.json()["commit"])

    def get_connected_storages(self) -> List[StorageAPIEntry]:
        """
        Get storages that are connected to the repository
        """
        res = self._http_request("GET", self.storage_api_url())

        if res.status_code == 404:
            raise RepoNotFoundError(f"Repo {self.repo_url} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting repository info."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        return [dacite.from_dict(StorageAPIEntry, storage_entry) for storage_entry in res.json()]

    def list_path(self, path: str, revision: Optional[str] = None, include_size: bool = False) -> List[ContentAPIEntry]:
        """
        List contents of a repository directory

        Args:
            path: Path of the directory
            revision: Branch or commit SHA. Default is tip of the default branch.
            include_size: Whether to include sizes for files. Calculating sizes might take more time.
        """
        params = {"include_size": include_size}
        res = self._http_request("GET", self.content_api_url(path, revision), params=params)

        if res.status_code == 404:
            raise PathNotFoundError(f"Path {path} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when listing path {path}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        content = res.json()
        if type(content) is dict:
            content = [content]
        return [dacite.from_dict(ContentAPIEntry, entry) for entry in content]

    def list_storage_path(self, path: str, include_size: bool = False) -> List[ContentAPIEntry]:
        """
        List contents of a folder in a connected storage bucket

        Args:
            path: Path of the storage directory in the format of ``<scheme>/<bucket-name>/<path>``

        Path example: ``s3/my-bucket/prefix/path/to/file``
        """
        params = {"include_size": include_size, "paging": True}

        url = self.storage_content_api_url(path)
        has_next_page = True

        def _get():
            res = self._http_request("GET", url, params=params)

            if res.status_code == 404:
                raise PathNotFoundError(f"Path {path} not found")
            elif res.status_code >= 400:
                error_msg = f"Got status code {res.status_code} when listing path {path}"
                logger.error(error_msg)
                logger.debug(res.content)
                raise RuntimeError(error_msg)

            content = res.json()
            if "entries" not in content:
                content = {
                    "entries": [
                        content,
                    ],
                    "next_token": None,
                }
            return dacite.from_dict(StorageContentAPIResult, content)

        entries = []

        while has_next_page:
            has_next_page = False
            resp = _get()
            entries += resp.entries
            if resp.next_token is not None:
                has_next_page = True
                params["from_token"] = resp.next_token

        return entries

    def get_file(self, path: str, revision: Optional[str] = None) -> bytes:
        """
        Download file from repo.

        Args:
            path: Path of the file in the repo.
            revision: Git branch or revision from which to download the file.

        Returns:
            bytes: The content of the file.
        """
        res = self._http_request("GET", self.raw_api_url(path, revision))
        if res.status_code == 404:
            raise PathNotFoundError(f"Path {path} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting file {path}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return res.content

    def get_storage_file(self, path: str) -> bytes:
        """
        Download file from a connected storage bucket.

        Args:
            path: Path in the bucket in the format of ``<scheme>/<bucket-name>/<path>``.

        Path example: ``s3/my-bucket/prefix/path/to/file``

        Returns:
            bytes: The content of the file.
        """
        res = self._http_request("GET", self.storage_raw_api_url(path))
        if res.status_code == 404:
            raise PathNotFoundError(f"Path {path} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting file {path}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return res.content

    def _get_files_in_path(
        self, path, revision=None, recursive=False, traverse_storages=False
    ) -> List[ContentAPIEntry]:
        """
        Walks through the path of the repo, returning non-dir entries
        """

        dir_queue = []
        files = []

        list_fn_folder = partial(self.list_path, revision=revision)
        list_fn_storage = self.list_storage_path

        def push_folder(folder_path):
            dir_queue.append((folder_path, list_fn_folder))

        def push_storage(storage_path):
            dir_queue.append((storage_path, list_fn_storage))

        # Initialize the queue
        path, is_storage_path = self._sanitize_storage_path(path)
        if is_storage_path:
            push_storage(path)
        else:
            push_folder(path)

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        task = progress.add_task("Traversing directories...", total=None)

        def step(step_path, list_fn):
            """
            step_path: path in the repo to list
            list_fn: which function to use to list it (can be list_path or list_storage_path)
            """
            res = list_fn(step_path)
            for entry in res:
                if entry.type == "file":
                    files.append(entry)
                elif recursive:
                    if entry.versioning == "bucket":
                        if traverse_storages:
                            push_storage(entry.path)
                    else:
                        push_folder(entry.path)
            progress.update(task, advance=1)

        with progress:
            while len(dir_queue):
                query_path, list_fn = dir_queue.pop(0)
                step(query_path, list_fn)

        return files

    def download(
        self,
        remote_path: Union[str, PathLike],
        local_path: Union[str, PathLike] = ".",
        revision: Optional[str] = None,
        recursive=True,
        keep_source_prefix=False,
        redownload=False,
        download_storages=False,
    ):
        """
        Downloads the contents of the repository at "remote_path" to the "local_path"

        Args:
            remote_path: Path in the repository of the folder or file to download.
            local_path: Where to download the files. Defaults to current working directory.
            revision: Repo revision or branch, if not specified - uses default repo branch.
                Ignored for downloading from buckets.
            recursive: Whether to download files recursively.
            keep_source_prefix: | Whether to keep the path of the folder in the download path or not.
                | Example: Given remote_path ``src/data`` and file ``test/file.txt``
                | if ``True``: will download to ``<local_path>/src/data/test/file.txt``
                | if ``False``: will download to ``<local_path>/test/file.txt``
            redownload: Whether to redownload files that already exist on the local filesystem.
                The downloader doesn't do any hash comparisons and only checks
                if a file already exists in the local filesystem or not.
            download_storages: If downloading the whole repo, by default we're not downloading the integrated storages
                Toggle this to ``True`` to change this behavior
        """
        traverse_storages = True
        if str(remote_path) == "/" and not download_storages:
            log_message(
                "Skipping downloading from connected storages. "
                "Set the `download_storages` flag if you want "
                "to download the whole content of the connected storages."
            )
            traverse_storages = False

        files = self._get_files_in_path(remote_path, revision, recursive, traverse_storages=traverse_storages)
        file_tuples = []
        if local_path is None:
            local_path = "."
        local_path = Path(local_path)
        # Strip the slashes from the beginning so the relative_to logic works
        remote_path = str(remote_path).lstrip("/")
        if not remote_path:
            remote_path = "/"
        # For storage paths get rid of the colon in the beginning of the schema, the download urls won't have it either
        remote_path, _ = self._sanitize_storage_path(remote_path)
        # Edge case - if the user requested a single file - different output path semantics
        if len(files) == 1 and files[0].path == remote_path:
            f = files[0]
            remote_path = PurePosixPath(f.path)
            # If local_path was specified, assume that the local_path is the exact name of the file
            if local_path != Path("."):
                # Saving to existing dir - append the name of remote file to the end a-la cp
                if local_path.is_dir():
                    remote_path = remote_path if keep_source_prefix else remote_path.name
                    file_path = local_path / remote_path
                else:
                    file_path = local_path
            else:
                file_path = remote_path if keep_source_prefix else remote_path.name
            file_tuples.append((f.download_url, file_path))
        else:
            for f in files:
                file_path_in_remote = PurePosixPath(f.path)
                remote_path_obj = PurePosixPath(remote_path)
                if not keep_source_prefix and remote_path != "/":
                    file_path = file_path_in_remote.relative_to(remote_path_obj)
                else:
                    file_path = file_path_in_remote
                file_path = local_path / file_path
                file_tuples.append((f.download_url, file_path))
        download_files(file_tuples, skip_if_exists=not redownload)
        log_message(f"Downloaded {len(files)} file(s) to {local_path.resolve()}")

    @staticmethod
    def _sanitize_storage_path(path: Union[str, PathLike]) -> Tuple[str, bool]:
        """
        Sanitizes storage paths for use in the traversal/download functions.
        When user asks for ``s3:/prefix/file``, we need to request ``storage/s3/prefix/file``.
        This function checks that the user has asked for a storage path (by the schema in the beginning)
        and returns a path that can be used for these types of requests.
        If the path is not a storage path, path is returned as is (converted to string).
        """
        path = str(path)
        if path.lstrip("/").split("/")[0] in {"s3:", "gs:", "azure:"}:
            path = str(path).lstrip("/").replace(":", "", 1)
            return path, True
        return path, False

    @cached_property
    def default_branch(self) -> str:
        """
        Name of the repository's default branch
        """
        return self.get_repo_info().default_branch

    @cached_property
    def id(self) -> int:
        return self.get_repo_info().id

    @cached_property
    def is_mirror(self) -> bool:
        return self.get_repo_info().mirror

    @property
    def full_name(self) -> str:
        """
        Full name of the repo in ``<owner>/<reponame>`` format
        """
        return f"{self.owner}/{self.repo_name}"

    def last_commit(self, branch: Optional[str] = None) -> CommitAPIResponse:
        """
        Returns info about the last commit of a branch.

        Args:
            branch: Branch to get the last commit of. Defaults to :func:`default_branch`.
        """
        if branch is None:
            branch = self.default_branch
        return self.get_branch_info(branch).commit

    def last_commit_sha(self, branch: Optional[str] = None) -> str:
        """
        Returns the SHA hash of the last commit of a branch.

        Args:
            branch: Branch to get the last commit of. Defaults to :func:`default_branch`.
        """
        return self.last_commit(branch).id

    @cached_property
    def repo_api_url(self) -> str:
        """
        Base URL for making all API request for the repos.
        Format: https://dagshub.com/api/v1/repos/user/repo

        :meta private:
        """
        return multi_urljoin(
            self.host,
            "api/v1/repos",
            self.owner,
            self.repo_name,
        )

    @cached_property
    def repo_url(self) -> str:
        """
        URL of the repo on DagsHub

        Format: ``https://dagshub.com/<user>/<repo>``
        """
        return multi_urljoin(self.host, self.owner, self.repo_name)

    def branch_url(self, branch) -> str:
        """
        URL of a branch on the repo
        Format: https://dasghub.com/api/v1/repos/user/repo/branches/branch

        :meta private:
        """
        return multi_urljoin(self.repo_api_url, "branches", branch)

    @cached_property
    def data_engine_url(self) -> str:
        """
        URL of data engine

        :meta private:
        """
        return multi_urljoin(self.repo_api_url, "data-engine")

    @cached_property
    def annotations_url(self) -> str:
        return multi_urljoin(self.repo_api_url, "annotations")

    def commit_url(self, sha) -> str:
        """
        URL of a commit in the repo
        Format: https://dagshub.com/api/v1/repos/user/repo/commits/sha

        :meta private:
        """
        return multi_urljoin(self.repo_api_url, "commits", sha)

    def content_api_url(self, path: str, revision: Optional[str] = None) -> str:
        """
        URL for Content API access
        Format: https://dagshub.com/api/v1/repos/user/repo/content/revision/path

        :meta private:
        """
        if revision is None:
            revision = self.default_branch
        return multi_urljoin(self.repo_api_url, "content", revision, path)

    def raw_api_url(self, path: str, revision: Optional[str] = None) -> str:
        """
        URL for Raw Content API access
        Format: https://dagshub.com/api/v1/repos/user/repo/raw/revision/path

        :meta private:
        """
        if revision is None:
            revision = self.default_branch
        return multi_urljoin(self.repo_api_url, "raw", revision, path)

    def storage_content_api_url(self, path: str) -> str:
        """
        URL for Storage Content API access
        path example: s3/bucket-name/path/in/bucket
        Format: https://dagshub.com/api/v1/repos/user/repo/storage/content/path

        :meta private:
        """
        return multi_urljoin(self.repo_api_url, "storage/content", path)

    def storage_raw_api_url(self, path: str) -> str:
        """
        URL for Storage Raw Content API access
        path example: s3/bucket-name/path/in/bucket
        Format: https://dagshub.com/api/v1/repos/user/repo/storage/raw/path

        :meta private:
        """
        return multi_urljoin(self.repo_api_url, "storage/raw", path)

    def storage_api_url(self) -> str:
        """
        URL for getting connected storages
        Format: https://dagshub.com/api/v1/repos/user/repo/storage

        :meta private:
        """
        return multi_urljoin(self.repo_api_url, "storage")

    def repo_bucket_api_url(self) -> str:
        """
        Endpoint URL for getting access to the S3-compatible repo bucket

        Format: https://dagshub.com/api/v1/repo-buckets/s3/user

        The bucket name is usually the name of the repo

        :meta private:
        """
        return multi_urljoin(self.host, "api/v1/repo-buckets/s3", self.owner)

    @staticmethod
    def parse_repo(repo: str) -> Tuple[str, str]:
        repo = repo.strip("/")
        parts = repo.split("/")
        if len(parts) != 2:
            raise WrongRepoFormatError("repo needs to be in the format <repo-owner>/<repo-name>")
        return parts[0], parts[1]



import errno
import logging
import os
import platform
import sys
from argparse import ArgumentParser
from os import PathLike
from pathlib import Path
from threading import Lock
from typing import Optional
from dagshub.common import rich_console

from .filesystem import SPECIAL_FILE, DagsHubFilesystem

logger = logging.getLogger(__name__)

SPECIAL_FILE_FH = (1 << 64) - 1

fuse_enabled_systems = ["Linux"]
system = platform.system()
if system not in fuse_enabled_systems:
    err_str = (
        f"FUSE mounting isn't supported on {system}.\n"
        f"Please use install_hooks to access DagsHub hosted files from a python script"
    )
    raise ImportError(err_str)
from fuse import FUSE, FuseOSError, LoggingMixIn, Operations  # noqa


class DagsHubFUSE(LoggingMixIn, Operations):
    def __init__(
        self,
        project_root: Optional[PathLike] = None,
        repo_url: Optional[str] = None,
        branch: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
    ):
        # FIXME TODO move autoconfiguration out of FUSE object constructor and to main method
        self.fs = DagsHubFilesystem(
            project_root=project_root,
            repo_url=repo_url,
            branch=branch,
            username=username,
            password=password,
            token=token,
        )
        logger.debug("__init__")
        self.rwlock = Lock()

    def __call__(self, op, path, *args):
        return super(DagsHubFUSE, self).__call__(op, self.fs.project_root / path[1:], *args)

    def access(self, path, mode):
        """
        Check file accessibility based on the given path and access mode.

        Args:
            path (Union[str, int, bytes]): The path to check accessibility.
                It can be a path (str), file descriptor (int), or bytes-like object.
            mode (int):
                The access mode to check.

        Returns:
            bool: True if the file is accessible; otherwise, False.

        Notes:
            - If the provided 'path' argument is an integer (file descriptor),
                the function behaves as a passthrough to the standard access() method.
            - The 'mode' argument follows the same convention as the os.access() function,
                where values like os.R_OK, os.W_OK, and os.X_OK indicate read, write, and execute permissions.

        Examples:
            ```python
            dh = DagsHubClient()
            is_accessible = dh.access('file.txt', os.R_OK)
            print(is_accessible)  # True if the file is readable, otherwise False
            ```
        """
        logger.debug(f"access - path: {path}, mode:{mode}")
        try:
            self.fs.stat(path)
        except FileNotFoundError:
            return False

    def open(self, path, flags):
        """
        NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/functions.html#open)

        Open a file for reading or writing.

        Args:
            path (Union[str, int, bytes]): The path of the file to open.
                It can be a path (str), file descriptor (int), or bytes-like object.
            flags (int): The flags for opening the file.

        Raises:
            FuseOSError: If an error occurs while opening the file, a FuseOSError is raised.

        Returns:
            int: The file descriptor for the opened file.

        """
        logger.debug(f"open - path: {path}, flags: {flags}")
        if path == Path(self.fs.project_root / SPECIAL_FILE):
            return SPECIAL_FILE_FH
        try:
            self.fs.open(path).close()
        except FileNotFoundError:
            raise FuseOSError(errno.ENOENT)
        logger.debug("finished fs.open")
        return os.open(self.fs._relative_path(path), flags, dir_fd=self.fs.project_root_fd)

    def getattr(self, path, fd=None):
        """
        NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/functions.html#getattr)

        Get the attributes of a file or directory.

        Args:
            path (Union[str, int, bytes]): The path to the file or directory.
                It can be a path (str), file descriptor (int), or bytes-like object.
            fd (int, optional): An optional file descriptor. Defaults to None.

        Raises:
            FuseOSError: If the file or directory does not exist, a FuseOSError is raised.

        """
        logger.debug(f"getattr - path:{str(path)}, fd:{fd}")
        try:
            if fd:
                logger.debug("with __stat")
                st = self.fs._DagsHubFilesystem__stat(fd)
            else:
                logger.debug("with fs.stat")
                st = self.fs.stat(path)

            logger.debug(f"st: {st}")
            return {
                key: getattr(st, key)
                for key in (
                    "st_atime",
                    "st_ctime",
                    "st_gid",
                    "st_mode",
                    "st_mtime",
                    # 'st_nlink',
                    "st_size",
                    "st_uid",
                )
            }
        except FileNotFoundError:
            logger.debug("FileNotFound")
            raise FuseOSError(errno.ENOENT)

    def read(self, path, size, offset, fh):
        """
         NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/os.html#os.read)

        Read data in the form of bytes from a file.

        Args:
            path (Union[str, int, bytes]): The path of the file to read. It can be a path (str),
                file descriptor (int), or bytes-like object.
            size (int): The size of data to read.
            offset (int): The offset in the file.
            fh (int): The file descriptor.

        """
        logger.debug(f"read - path: {path}, offset: {offset}, fh: {fh}")
        if fh == SPECIAL_FILE_FH:
            return self.fs._special_file()[offset : offset + size]
        with self.rwlock:
            os.lseek(fh, offset, 0)
            return os.read(fh, size)

    def readdir(self, path, fh):
        """
        List the contents of a directory.

        Args:
            path (Union[str, int, bytes]): The path of the directory.
                It can be a path (str), file descriptor (int), or bytes-like object.
            fh (int): The file descriptor.

        Returns:
            List[str]: A list of directory contents.

        """
        logger.debug(f"readdir - path: {path}, fh: {fh}")
        return [".", ".."] + self.fs.listdir(path)

    def release(self, path, fh):
        """
        Release the resources associated with an open file.

        Args:
            path (Union[str, int, bytes]):
                The path of the file.
                It can be a path (str), file descriptor (int), or bytes-like object.
            fh (int):
                The file descriptor.

        Notes:
            - If the provided 'path' argument is an integer (file descriptor),
                the function behaves as a passthrough to the standard os.close() method.
            - Special file descriptors, such as SPECIAL_FILE_FH, are not closed.

        Examples:
            ```python
            dh = DagsHubClient()
            dh.release('file.txt', file_descriptor)
            ```
        """
        logger.debug(f"release - path: {path}, fh: {fh}")
        if fh != SPECIAL_FILE_FH:
            return os.close(fh)


def mount(
    debug=False,
    project_root: Optional[PathLike] = None,
    repo_url: Optional[str] = None,
    branch: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    token: Optional[str] = None,
):
    """
    Mount a DagsHubFUSE filesystem.

    Args:
        debug (bool, optional): If True, run the FUSE filesystem in the foreground with debug logging;
            otherwise, run it in the background. Defaults to False.
        project_root (Optional[PathLike], optional): The local directory to mount as the DagsHubFUSE filesystem.
            Defaults to None.
        repo_url (Optional[str], optional): The URL of the DagsHub repository to mount. Defaults to None.
        branch (Optional[str], optional): The branch of the DagsHub repository to mount. Defaults to None.
        username (Optional[str], optional): The username for authentication. Defaults to None.
        password (Optional[str], optional): The password for authentication. Defaults to None.
        token (Optional[str], optional): The token for authentication. Defaults to None.

    Notes:
        - If the 'debug' parameter is True, the filesystem is run in the foreground with debug logging.
        - If 'debug' is False, the filesystem runs in the background.

    Example:
        ```python
        mount(debug=True, project_root='/path/local/dir', repo_url='https://dagshub.com/user/repo.git', branch='main')
        ```
    """
    logging.basicConfig(level=logging.DEBUG)
    fuse = DagsHubFUSE(
        project_root=project_root, repo_url=repo_url, branch=branch, username=username, password=password, token=token
    )
    rich_console.print(
        f"Mounting DagsHubFUSE filesystem at {fuse.fs.project_root}\n"
        f"Run `cd .` in any existing terminals to utilize mounted FS."
    )
    FUSE(fuse, str(fuse.fs.project_root), foreground=debug, nonempty=True)
    if not debug:
        os.chdir(os.path.realpath(os.curdir))
    # TODO: Clean unmounting procedure


def main():
    parser = ArgumentParser()
    parser.add_argument("project_root", nargs="?")
    parser.add_argument("--repo_url")
    parser.add_argument("--branch")
    parser.add_argument("--username")
    parser.add_argument("--password")
    parser.add_argument("--debug", action="store_true", default=False)  # default=False, nargs=0)

    args = parser.parse_args()

    if not args.debug:
        # Hide tracebacks of errors, display only error message
        sys.tracebacklimit = 0

    mount(**vars(args))


if __name__ == "__main__":
    main()



from pathlib import Path


class DagsHubRepoNotFoundError(Exception):
    def __init__(self, path: Path):
        super().__init__()
        self.path = path

    def __str__(self):
        return f"Couldn't find a DagsHub repo in the path {self.path} or its parents"



from pathlib import Path


class FilesystemAlreadyMountedError(Exception):
    def __init__(self, path: Path, repo: str, revision: str):
        self.path = path
        self.repo = repo
        self.revision = revision

    def __str__(self):
        return (
            f"There is already a filesystem mounted at path {self.path.absolute()} "
            f"({self.repo} revision {self.revision})"
            f"\ndel() the filesystem object in use if you want to switch the mounted filesystem"
        )



import hashlib
from typing import Optional
import logging
import urllib
import uuid
import httpx
import webbrowser

from dagshub.auth.token_auth import OAuthDagshubToken
from dagshub.common import config, rich_console

logger = logging.getLogger(__name__)


def oauth_flow(host: str, client_id: Optional[str] = None) -> OAuthDagshubToken:
    """
    Initiate the OAuth 2.0 flow for obtaining an access token.

    Args:
        host (str): The URL of the OAuth provider.
        client_id (Optional[str], optional): The client ID used for authentication.
            If not provided, it will use the default client ID from the configuration.

    Returns:
        Dict: A dictionary containing the obtained access token.
    """

    host = host.strip("/")
    dagshub_url = urllib.parse.urljoin(host, "login/oauth")
    client_id = client_id or config.client_id
    state = uuid.uuid4()
    middle_man_request_id = hashlib.sha256(uuid.uuid4().bytes).hexdigest()
    auth_link = (
        f"{dagshub_url}/authorize?state={state}&client_id={client_id}" f"&middleman_request_id={middle_man_request_id}"
    )

    webbrowser.open(auth_link)

    rich_console.print(
        "[bold]:exclamation::exclamation::exclamation: AUTHORIZATION REQUIRED "
        ":exclamation::exclamation::exclamation:[/bold]",
        justify="center",
    )
    # Doing raw prints here because the rich syntax breaks in colab
    # Printing them line by line, because the link has to be its own print in order for Colab parser to correctly parse
    # the whole link
    print("\n\nOpen the following link in your browser to authorize the client:")
    print(auth_link)
    print("\n")

    with rich_console.status("Waiting for authorization"):
        res = httpx.post(f"{dagshub_url}/middleman", data={"request_id": middle_man_request_id}, timeout=None)

    if res.status_code != 200:
        raise Exception(f"Error while getting OAuth code: HTTP {res.status_code}, Body: {res.json()}")
    code = res.json()

    res = httpx.post(f"{dagshub_url}/access_token", data={"client_id": client_id, "code": code, "state": str(state)})

    if res.status_code != 200:
        raise Exception(f"Error while getting OAuth token: HTTP {res.status_code}, Body: {res.json()}")
    token = OAuthDagshubToken.deserialize(res.json())

    logger.debug(f"Got token: {token}")
    return token


class OauthNonInteractiveShellException(Exception):
    pass



import base64
import datetime
import gzip
import json
import logging
import math
import os.path
import time
import webbrowser
from contextlib import contextmanager
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union, Set, ContextManager, Tuple

import rich.progress
from dataclasses_json import config, LetterCase, DataClassJsonMixin
from pathvalidate import sanitize_filepath

import dagshub.common.config
from dagshub.common import rich_console
from dagshub.common.analytics import send_analytics_event
from dagshub.common.helpers import prompt_user, http_request, log_message
from dagshub.common.rich_util import get_rich_progress
from dagshub.common.util import lazy_load, multi_urljoin, to_timestamp, exclude_if_none
from dagshub.data_engine.client.models import (
    PreprocessingStatus,
    MetadataFieldSchema,
    ScanOption,
    autogenerated_columns,
    DatasetResult,
)
from dagshub.data_engine.dtypes import MetadataFieldType
from dagshub.data_engine.model.datapoint import Datapoint
from dagshub.data_engine.model.errors import (
    WrongOperatorError,
    WrongOrderError,
    DatasetFieldComparisonError,
    FieldNotFoundError,
    DatasetNotFoundError,
)
from dagshub.data_engine.model.metadata_field_builder import MetadataFieldBuilder
from dagshub.data_engine.model.query import QueryFilterTree
from dagshub.data_engine.model.schema_util import metadataTypeLookup, metadataTypeLookupReverse
from dagshub.data_engine.model.datasource_state import DatasourceState

if TYPE_CHECKING:
    from dagshub.data_engine.model.query_result import QueryResult
    import fiftyone as fo
    import pandas
    import mlflow
    import mlflow.entities
else:
    plugin_server_module = lazy_load("dagshub.data_engine.voxel_plugin_server.server")
    fo = lazy_load("fiftyone")
    mlflow = lazy_load("mlflow")

logger = logging.getLogger(__name__)

DEFAULT_MLFLOW_ARTIFACT_NAME = "datasource.dagshub.json"


@dataclass
class DatapointMetadataUpdateEntry(DataClassJsonMixin):
    url: str
    key: str
    value: str
    valueType: MetadataFieldType = field(metadata=config(encoder=lambda val: val.value))
    allowMultiple: bool = False


@dataclass
class DatapointDeleteMetadataEntry(DataClassJsonMixin):
    datapointId: str
    key: str


@dataclass
class DatapointDeleteEntry(DataClassJsonMixin):
    datapointId: str


@dataclass
class Field:
    """
    Class used to define custom fields for use in \
    :func:`Datasource.select() <dagshub.data_engine.model.datasource.Datasource.select>` or in filtering.

    Example of filtering on old data from a field::

        t = datetime.now() - timedelta(days=2)
        q = ds[Field("size", as_of=t)] > 500
        q.all()
    """

    field_name: str
    """The database field where the values are stored. In other words, where to get the values from."""
    as_of: Optional[Union[float, datetime.datetime]] = None
    """
    If defined, the data in this field would be shown as of this moment in time.

    Accepts either a datetime object, or a UTC timestamp.
    """
    alias: Optional[str] = None
    """
    How the returned custom data field should be named.

    Useful when you're comparing the same field at multiple points in time::

        yesterday = datetime.now() - timedelta(days=1)

        ds.select(
            Field("value", alias="value_today"),
            Field("value", as_of=yesterday, alias="value_yesterday")
        ).all()
    """

    @property
    def as_of_timestamp(self) -> Optional[int]:
        if self.as_of is None:
            return None
        return to_timestamp(self.as_of)

    def to_dict(self, ds: "Datasource") -> Dict[str, Any]:
        if not ds.has_field(self.field_name):
            raise FieldNotFoundError(self.field_name)

        res_dict: Dict[str, Union[str, int, None]] = {"name": self.field_name}
        if self.as_of is not None:
            res_dict["asOf"] = self.as_of_timestamp
        if self.alias:
            res_dict["alias"] = self.alias
        return res_dict


class Datasource:
    def __init__(
        self,
        datasource: "DatasourceState",
        query: Optional["DatasourceQuery"] = None,
        from_dataset: Optional["DatasetState"] = None,
    ):
        self._source = datasource
        if query is None:
            query = DatasourceQuery()
        self._query = query
        # a per datasource context used for dict update syntax
        self._implicit_update_ctx: Optional[MetadataContextManager] = None
        # this ref marks if source is currently used in
        # meta-data update 'with' block
        self._explicit_update_ctx: Optional[MetadataContextManager] = None
        self.assigned_dataset = from_dataset

    @property
    def has_explicit_context(self):
        return self._explicit_update_ctx is not None

    def _get_source_implicit_metadata_entries(self):
        if self._implicit_update_ctx is not None:
            return self._implicit_update_ctx.get_metadata_entries()
        else:
            return []

    @property
    def source(self) -> "DatasourceState":
        return self._source

    def clear_query(self, reset_to_dataset=True):
        """
        Clear the attached query.

        Args:
            reset_to_dataset: If ``True`` and this Datasource was saved as a dataset, reset to the query in the dataset,
                instead of clearing the query completely.
        """
        if reset_to_dataset and self.assigned_dataset is not None and self.assigned_dataset.query is not None:
            self._query = self.assigned_dataset.query.__deepcopy__()
        else:
            self._query = DatasourceQuery()

    def __deepcopy__(self, memodict={}) -> "Datasource":
        res = Datasource(self._source, self._query.__deepcopy__())
        res.assigned_dataset = self.assigned_dataset

        # Carry over the update context, that way we'll keep track of the metadata being uploaded
        res._implicit_update_ctx = self._implicit_update_ctx

        return res

    def get_query(self) -> "DatasourceQuery":
        return self._query

    @property
    def annotation_fields(self) -> List[str]:
        """Return all fields that have the annotation meta tag set"""
        return [f.name for f in self.fields if f.is_annotation()]

    def serialize_gql_query_input(self) -> Dict:
        """
        Serialize the query of this Datasource for use in GraphQL querying (e.g. getting datapoints)

        :meta private:
        """
        return self._query.to_dict()

    def _deserialize_from_gql_result(self, query_dict: Dict):
        """
        Imports query information from ``query_dict``
        """
        self._query = DatasourceQuery.from_dict(query_dict)

    def load_from_dataset(
        self, dataset_id: Optional[Union[str, int]] = None, dataset_name: Optional[str] = None, change_query=True
    ):
        """
        Imports query information from a dataset with the specified id or name. Either of id or name could be specified

        Args:
            dataset_id: ID of the dataset
            dataset_name: Name of the dataset
            change_query: Whether to change the query of this object to the query in the dataset

        :meta private:
        """
        datasets = self.source.client.get_datasets(id=dataset_id, name=dataset_name)
        if not datasets:
            raise DatasetNotFoundError(self.source.repo, dataset_id, dataset_name)
        dataset_state = DatasetState.from_gql_dataset_result(datasets[0])
        self.load_from_dataset_state(dataset_state, change_query)

    def load_from_dataset_state(self, dataset_state: "DatasetState", change_query=True):
        """
        Imports query information from a :class:`~dagshub.data_engine.model.dataset_state.DatasetState`

        Args:
            dataset_state: State to load
            change_query: If false, only assigns the dataset.
                If true, also changes the query to be the query of the dataset

        :meta private:
        """
        if self.source.id != dataset_state.datasource_id:
            raise RuntimeError(
                "Dataset belongs to a different datasource "
                f"(This datasource: {self.source.id}, Dataset's datasource: {dataset_state.datasource_id})"
            )
        self.assigned_dataset = dataset_state
        if change_query:
            self._query = dataset_state.query

    def sample(self, start: Optional[int] = None, end: Optional[int] = None):
        if start is not None:
            logger.warning("Starting slices is not implemented for now")
        return self._source.client.sample(self, end, include_metadata=True)

    def head(self, size=100) -> "QueryResult":
        """
        Executes the query and returns a :class:`.QueryResult` object containing first ``size`` datapoints

        Args:
            size: how many datapoints to get. Default is 100
        """
        self._check_preprocess()
        send_analytics_event("Client_DataEngine_DisplayTopResults", repo=self.source.repoApi)
        return self._source.client.head(self, size)

    def all(self) -> "QueryResult":
        """
        Executes the query and returns a :class:`.QueryResult` object containing all datapoints
        """
        self._check_preprocess()
        return self._source.client.get_datapoints(self)

    def select(self, *selected: Union[str, Field]) -> "Datasource":
        """
        Select which fields should appear in the query result.

        If you want to query older versions of metadata,
        use :class:`Field` objects with ``as_of`` set to your desired time.

        By default, only the defined fields are returned.
        If you want to return all existing fields plus whatever additional fields you define,
        add ``"*"`` into the arguments.

        Args:
            selected: Fields you want to select. Can be either of:

                - Name of the field to select: ``"field"``.
                - ``"*"`` to select all the fields in the datasource.
                - :class:`Field` object.

        Example::

            t = datetime.now() - timedelta(hours=24)
            q1 = ds.select("*", Field("size", as_of=t, alias="size_asof_24h_ago"))
            q1.all()
        """

        include_all = False
        selects = []
        for s in selected:
            if isinstance(s, Field):
                selects.append(s.to_dict(self))
            else:
                if s != "*":
                    selects.append({"name": s})
                else:
                    include_all = True

        if include_all:
            aliases = set([s["alias"] for s in selects if "alias" in s])
            for f in self.fields:
                if f.name in aliases:
                    raise ValueError(
                        f"Alias {f.name} can't be used, because * was specified and "
                        f"a field with that name already exists"
                    )
                selects.append({"name": f.name})

        new_ds = self.__deepcopy__()
        new_ds.get_query().select = selects
        return new_ds

    def as_of(self, time: Union[float, datetime.datetime]) -> "Datasource":
        """
        Get a snapshot of the datasource's state as of ``time``.

        Args:
            time: At which point in time do you want to get data from.\
                Either a UTC timestamp or a ``datetime`` object.

        In the following example, you will get back datapoints that were created no later than yesterday AND \
        had their size at this point bigger than 5 bytes::

            t = datetime.now() - timedelta(hours=24)
            q1 = (ds["size"] > 5).as_of(t)
            q1.all()

        .. note::
            If used with :func:`select`, the ``as_of`` set on the fields takes precedence
            over the global query ``as_of`` set here.
        """
        new_ds = self.__deepcopy__()

        new_ds._query.as_of = to_timestamp(time)
        return new_ds

    def order_by(self, *args: Union[str, Tuple[str, Union[bool, str]]]) -> "Datasource":
        """
        Sort the query result by the specified fields.
        Any previously set order will be overwritten.

        Args:
            Fields to sort by. Can be either of:
                - Name of the field to sort by: ``"field"``.
                - A tuple of ``(field_name, ascending)``: ``("field", True)``.
                - A tuple of ``(field_name, "asc"|"desc")``: ``("field", "asc")``.

        Examples::

            ds.order_by("size").all()                   # Order by ascending size
            ds.order_by(("date", "desc"), "size).all()  # Order by descending date, then ascending size
        """
        new_ds = self.__deepcopy__()
        orders = []
        for arg in args:
            if isinstance(arg, str):
                orders.append({"field": arg, "order": "ASC"})
            else:
                if len(arg) != 2:
                    raise RuntimeError(
                        f"Invalid sort argument {arg}, must be a tuple (<field>, 'asc'|'desc'|True|False)"
                    )
                if isinstance(arg[1], bool):
                    order = "ASC" if arg[1] else "DESC"
                elif isinstance(arg[1], str) and arg[1].upper() in ["ASC", "DESC"]:
                    order = arg[1].upper()
                else:
                    raise RuntimeError(f"Invalid sort argument {arg}, second value must be 'asc'|'desc'|True|False")
                orders.append({"field": arg[0], "order": order})
        new_ds.get_query().order_by = orders
        return new_ds

    def _check_preprocess(self):
        self.source.get_from_dagshub()
        if (
            self.source.preprocessing_status == PreprocessingStatus.IN_PROGRESS
            or self.source.preprocessing_status == PreprocessingStatus.AUTO_SCAN_IN_PROGRESS
        ):
            logger.warning(
                f"Datasource {self.source.name} is currently in the progress of rescanning. "
                f"Values might change if you requery later"
            )

    def metadata_field(self, field_name: str) -> MetadataFieldBuilder:
        """
        Returns a builder for a metadata field.
        The builder can be used to change properties of a field or create a new field altogether.
        Note that fields get automatically created when you upload new metadata to the Data Engine,
        so it's not necessary to create fields with this function.

        Example of creating a new annotation field::

            ds.metadata_field("annotation").set_type(dtypes.LabelStudioAnnotation).apply()

        .. note::
            New fields have to have their type defined using ``.set_type()`` before doing anything else

        Example of marking an existing field as an annotation field::

            ds.metadata_field("existing-field").set_annotation().apply()

        Args:
            field_name: Name of the field that you want to create/change

        """
        return MetadataFieldBuilder(self, field_name)

    def apply_field_changes(self, field_builders: List[MetadataFieldBuilder]):
        """
        Applies one or multiple metadata field builders
        that can be constructed using the :func:`metadata_field()` function.
        """
        self.source.client.update_metadata_fields(self, [builder.schema for builder in field_builders])
        self.source.get_from_dagshub()

    @property
    def implicit_update_context(self) -> "MetadataContextManager":
        if not self._implicit_update_ctx:
            self._implicit_update_ctx = MetadataContextManager(self)

        return self._implicit_update_ctx

    def upload_metadata_of_implicit_context(self):
        """
        commit meta data changes done in dictionary assignment context
        :meta private:
        """
        if self._implicit_update_ctx:
            try:
                self._upload_metadata(self._get_source_implicit_metadata_entries())
            finally:
                self._implicit_update_ctx = None

    def metadata_context(self) -> ContextManager["MetadataContextManager"]:
        """
        Returns a metadata context, that you can upload metadata through using its
        :func:`~MetadataContextManager.update_metadata()` function.
        Once the context is exited, all metadata is uploaded in one batch::

            with df.metadata_context() as ctx:
                ctx.update_metadata("file1", {"key1": True, "key2": "value"})

        """

        # Need to have the context manager inside a wrapper to satisfy MyPy + PyCharm type hinter
        @contextmanager
        def func():
            self.source.get_from_dagshub()
            send_analytics_event("Client_DataEngine_addEnrichments", repo=self.source.repoApi)
            ctx = MetadataContextManager(self)

            self._explicit_update_ctx = ctx
            yield ctx
            try:
                self._upload_metadata(ctx.get_metadata_entries() + self._get_source_implicit_metadata_entries())
            finally:
                self._implicit_update_ctx = None
                self._explicit_update_ctx = None

        return func()

    def upload_metadata_from_dataframe(self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None):
        """
        Upload metadata from a pandas dataframe.

        All columns are uploaded as metadata, and the path of every datapoint is taken from `path_column`.

        Args:
            df (`pandas.DataFrame <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.html>`_):
                DataFrame with metadata
            path_column: Column with the datapoints' paths. Can either be the name of the column, or its index.
                If not specified, the first column is used.
        """
        self.source.get_from_dagshub()
        send_analytics_event("Client_DataEngine_addEnrichmentsWithDataFrame", repo=self.source.repoApi)
        self._upload_metadata(self._df_to_metadata(df, path_column, multivalue_fields=self._get_multivalue_fields()))

    def _get_multivalue_fields(self) -> Set[str]:
        res = set()
        for col in self.source.metadata_fields:
            if col.multiple:
                res.add(col.name)
        return res

    def _df_to_metadata(
        self, df: "pandas.DataFrame", path_column: Optional[Union[str, int]] = None, multivalue_fields=set()
    ) -> List[DatapointMetadataUpdateEntry]:
        res: List[DatapointMetadataUpdateEntry] = []
        if path_column is None:
            path_column = df.columns[0]
        elif type(path_column) is str:
            if path_column not in df.columns:
                raise RuntimeError(f"Column {path_column} does not exist in the dataframe")
        elif type(path_column) is int:
            path_column = df.columns[path_column]

        # objects are actually mixed and not guaranteed to be string, but this should cover most use cases
        if df.dtypes[path_column] != "object":
            raise RuntimeError(f"Column {path_column} doesn't have strings")

        field_value_types = {f.name: f.valueType for f in self.fields}

        for _, row in df.iterrows():
            datapoint = row[path_column]
            for key, val in row.items():
                if key == path_column:
                    continue
                key = str(key)
                if key in autogenerated_columns:
                    continue
                if val is None:
                    continue
                # ONLY FOR PANDAS: since pandas doesn't distinguish between None and NaN, don't upload it
                if type(val) is float and math.isnan(val):
                    continue
                if type(val) is list:
                    if key not in multivalue_fields:
                        multivalue_fields.add(key)
                        # Promote all the existing uploading metadata to multivalue
                        for update_entry in res:
                            if update_entry.key == key:
                                update_entry.allowMultiple = True
                    for sub_val in val:
                        value_type = field_value_types.get(key)
                        if value_type is None:
                            value_type = metadataTypeLookup[type(sub_val)]
                            field_value_types[key] = value_type
                        # Don't override bytes if they're not bytes - probably just undownloaded values
                        if value_type == MetadataFieldType.BLOB and type(sub_val) is not bytes:
                            continue
                        # Pandas quirk - integers are floats on the backend
                        if value_type == MetadataFieldType.INTEGER:
                            sub_val = int(sub_val)
                        if type(sub_val) is bytes:
                            sub_val = MetadataContextManager.wrap_bytes(sub_val)
                        res.append(
                            DatapointMetadataUpdateEntry(
                                url=datapoint, key=key, value=str(sub_val), valueType=value_type, allowMultiple=True
                            )
                        )
                else:
                    value_type = field_value_types.get(key)
                    if value_type is None:
                        value_type = metadataTypeLookup[type(val)]
                        field_value_types[key] = value_type
                    # Don't override bytes if they're not bytes - probably just undownloaded values
                    if value_type == MetadataFieldType.BLOB and type(val) is not bytes:
                        continue
                    # Pandas quirk - integers are floats on the backend
                    if value_type == MetadataFieldType.INTEGER:
                        val = int(val)
                    if type(val) is bytes:
                        val = MetadataContextManager.wrap_bytes(val)
                    res.append(
                        DatapointMetadataUpdateEntry(
                            url=datapoint,
                            key=key,
                            value=str(val),
                            valueType=value_type,
                            allowMultiple=key in multivalue_fields,
                        )
                    )
        return res

    def delete_source(self, force: bool = False):
        """
        Delete the record of this datasource along with all datapoints.

        .. warning::
            This is a destructive operation! If you delete the datasource,
            all the datapoints and metadata will be removed.

        Args:
            force: Skip the confirmation prompt
        """
        prompt = (
            f'You are about to delete datasource "{self.source.name}" for repo "{self.source.repo}"\n'
            f"This will remove the datasource and ALL datapoints "
            f"and metadata records associated with the source."
        )
        if not force:
            user_response = prompt_user(prompt)
            if not user_response:
                print("Deletion cancelled")
                return
        self.source.client.delete_datasource(self)

    def scan_source(self, options: Optional[List[ScanOption]] = None):
        """
        This function fires a call to the backend to rescan the datapoints.
        Call this function whenever you uploaded new files and want them to appear when querying the datasource,
        or if you changed existing file contents and want their metadata to be updated.

        DagsHub periodically rescans all datasources, this function is a way to make a scan happen as soon as possible.

        Notes about automatically scanned metadata:
            1. Only new datapoints (files) will be added.
               If files were removed from the source, their metadata will still remain,
               and they will still be returned from queries on the datasource.
               An API to actively remove metadata will be available soon.
            2. Some metadata fields will be automatically scanned and updated by DagsHub based on this scan -
               the list of automatic metadata fields is growing frequently!

        Args:
            options: List of scanning options. If not sure, leave empty.
        """
        logger.debug("Rescanning datasource")
        self.source.client.scan_datasource(self, options=options)

    def _upload_metadata(self, metadata_entries: List[DatapointMetadataUpdateEntry]):
        progress = get_rich_progress(rich.progress.MofNCompleteColumn())

        upload_batch_size = dagshub.common.config.dataengine_metadata_upload_batch_size
        total_entries = len(metadata_entries)
        total_task = progress.add_task(f"Uploading metadata (batch size {upload_batch_size})...", total=total_entries)

        with progress:
            for start in range(0, total_entries, upload_batch_size):
                entries = metadata_entries[start : start + upload_batch_size]
                logger.debug(f"Uploading {len(entries)} metadata entries...")
                self.source.client.update_metadata(self, entries)
                progress.update(total_task, advance=upload_batch_size)
            progress.update(total_task, completed=total_entries, refresh=True)

        # Update the status from dagshub, so we get back the new metadata columns
        self.source.get_from_dagshub()

    def delete_metadata_from_datapoints(self, datapoints: List[Datapoint], fields: List[str]):
        """
        Delete metadata from datapoints.
        The deleted values can be accessed using versioned query with time set before the deletion

        Args:
            datapoints: datapoints to delete metadata from
            fields: fields to delete
        """
        metadata_entries = []
        for d in datapoints:
            for n in fields:
                metadata_entries.append(DatapointDeleteMetadataEntry(datapointId=d.datapoint_id, key=n))
        self.source.client.delete_metadata_for_datapoint(self, metadata_entries)

    def delete_datapoints(self, datapoints: List[Datapoint], force: bool = False):
        """
        Delete datapoints.

        - These datapoints will no longer show up in queries.
        - Does not delete the datapoint's file, only removing the data from the datasource.
        - You can still query these datapoints and associated metadata with \
        versioned queries whose time is before deletion time.
        - You can re-add these datapoints to the datasource by uploading new metadata to it with, for example, \
        :func:`Datasource.metadata_context <dagshub.data_engine.model.datasource.Datasource.metadata_context>`. \
        This will create a new datapoint with new id and new metadata records.
        - Datasource scanning will *not* add these datapoints back.

        Args:
            datapoints: list of datapoints objects to delete
            force: Skip the confirmation prompt
        """
        dps_str = "\n\t".join([""] + [d.path for d in datapoints])
        prompt = (
            f"You are about to delete the following datapoint(s): {dps_str}\n"
            f"This will remove the datapoint and metadata from unversioned queries, "
            f"but won't delete the underlying file."
        )
        if not force:
            user_response = prompt_user(prompt)
            if not user_response:
                print("Deletion cancelled")
                return

        self.source.client.delete_datapoints(
            self, [DatapointDeleteEntry(datapointId=d.datapoint_id) for d in datapoints]
        )

    def save_dataset(self, name: str) -> "Datasource":
        """
        Save the dataset, which is a combination of datasource + query, on the backend.
        That way you can persist and share your queries.
        You can get the dataset back later by calling :func:`.datasets.get_dataset()`

        Args:
            name: Name of the dataset

        Returns:
            A datasource object with the dataset assigned to it
        """
        send_analytics_event("Client_DataEngine_QuerySaved", repo=self.source.repoApi)

        self.source.client.save_dataset(self, name)
        log_message(f"Dataset {name} saved")

        copy_with_ds_assigned = self.__deepcopy__()
        copy_with_ds_assigned.load_from_dataset(dataset_name=name, change_query=False)
        return copy_with_ds_assigned

    def log_to_mlflow(
        self, artifact_name=DEFAULT_MLFLOW_ARTIFACT_NAME, run: Optional["mlflow.entities.Run"] = None
    ) -> "mlflow.Entities.Run":
        """
        Logs the current datasource state to MLflow as an artifact.

        Args:
            artifact_name: Name of the artifact that will be stored in the MLflow run.
            run: MLflow run to save to. If ``None``, uses the active MLflow run or creates a new run.

        Returns:
            Run to which the artifact was logged.
        """
        if run is None:
            run = mlflow.active_run()
            if run is None:
                run = mlflow.start_run()
        mlflow.MlflowClient().log_dict(run.info.run_id, self._to_dict(), artifact_name)
        log_message(f'Saved the datasource state to MLflow (run "{run.info.run_name}") as "{artifact_name}"')
        return run

    def save_to_file(self, path: Union[str, PathLike] = ".") -> Path:
        """
        Saves a JSON file representing the current state of datasource or dataset.
        Useful for connecting code versions to the datasource used for training.

        .. note::
            Does not save the actual contents of the datasource/dataset, only the query.

        Args:
            path: Where to save the file. If path is an existing folder, saves to ``<path>/<ds_name>.json``.

        Returns:
            The path to the saved file
        """
        path = Path(path)
        if path.is_dir():
            if self.assigned_dataset is not None:
                name = self.assigned_dataset.dataset_name
            else:
                name = self.source.name
            path = path / f"{name}.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        res = self._to_dict()
        with open(path, "w") as file:
            json.dump(res, file, indent=4, sort_keys=True)
        log_message(f"Datasource saved to '{path}'")

        return path

    def _serialize(self) -> "DatasourceSerializedState":
        res = DatasourceSerializedState(
            repo=self.source.repo,
            datasource_id=self.source.id,
            datasource_name=self.source.name,
            query=self._query,
            timestamp=datetime.datetime.now().timestamp(),
            modified=self.is_query_different_from_dataset,
        )
        if self.assigned_dataset is not None:
            res.dataset_id = self.assigned_dataset.dataset_id
            res.dataset_name = self.assigned_dataset.dataset_name
        return res

    def _to_dict(self) -> Dict:
        res = self._serialize().to_dict()
        # Skip Nones in the result
        res = {k: v for k, v in res.items() if v is not None}
        return res

    @property
    def is_query_different_from_dataset(self) -> Optional[bool]:
        """
        Is the current query of the object different from the one in the assigned dataset.

        If no dataset is assigned, returns ``None``.
        """
        if self.assigned_dataset is None:
            return None
        return self._query.to_dict() != self.assigned_dataset.query.to_dict()

    @staticmethod
    def load_from_serialized_state(state_dict: Dict) -> "Datasource":
        """
        Load a Datasource that was saved with :func:`save_to_file`

        Args:
            state_dict: Serialized JSON object
        """

        state = DatasourceSerializedState.from_dict(state_dict)

        ds_state = DatasourceState(repo=state.repo, name=state.datasource_name, id=state.datasource_id)
        ds_state.get_from_dagshub()
        ds = Datasource(ds_state)
        ds._query = state.query

        if state.dataset_id is not None:
            ds.load_from_dataset(state.dataset_id, state.dataset_name, change_query=False)

        return ds

    def to_voxel51_dataset(self, **kwargs) -> "fo.Dataset":
        """
        Refer to :func:`QueryResult.to_voxel51_dataset() \
        <dagshub.data_engine.model.query_result.QueryResult.to_voxel51_dataset>`\
        for documentation.
        """
        return self.all().to_voxel51_dataset(**kwargs)

    @property
    def default_dataset_location(self) -> Path:
        """
        Default location where datapoint files are stored.

        On UNIX-likes the path is ``~/dagshub/datasets/<repo_name>/<datasource_id>``

        On Windows the path is ``C:\\Users\\<user>\\dagshub\\datasets\\<repo_name>\\<datasource_id>``
        """
        return Path(
            sanitize_filepath(os.path.join(Path.home(), "dagshub", "datasets", self.source.repo, str(self.source.id)))
        )

    def visualize(self, **kwargs) -> "fo.Session":
        """
        Visualize the whole datasource using
        :func:`QueryResult.visualize() <dagshub.data_engine.model.query_result.QueryResult.visualize>`.

        Read the function docs for kwarg documentation.
        """
        return self.all().visualize(**kwargs)

    @property
    def fields(self) -> List[MetadataFieldSchema]:
        return self.source.metadata_fields

    def annotate(self, fields_to_embed=None, fields_to_exclude=None) -> Optional[str]:
        """
        Sends all datapoints in the datasource for annotation in Label Studio.

        Args:
            fields_to_embed: list of meta-data columns that will show up in Label Studio UI.
             if not specified all will be displayed.
            fields_to_exclude: list of meta-data columns that will not show up in Label Studio UI

        .. note::
            This will send ALL datapoints in the datasource for annotation.
            It's recommended to not send a huge amount of datapoints to be annotated at once, to avoid overloading
            the Label Studio workspace.
            Use :func:`QueryResult.annotate() <dagshub.data_engine.model.query_result.QueryResult.annotate>`
            to annotate a result of a query with less datapoints.
            Alternatively, use a lower level :func:`send_datapoints_to_annotation` function

        :return: Link to open Label Studio in the browser
        """
        return self.all().annotate(fields_to_exclude=fields_to_exclude, fields_to_embed=fields_to_embed)

    def send_to_annotation(self):
        """
        deprecated, see :func:`annotate()`

        :meta private:
        """
        return self.annotate()

    def send_datapoints_to_annotation(
        self,
        datapoints: Union[List[Datapoint], "QueryResult", List[Dict]],
        open_project=True,
        ignore_warning=False,
        fields_to_exclude=None,
        fields_to_embed=None,
    ) -> Optional[str]:
        """
        Sends datapoints for annotation in Label Studio.

        Args:
            datapoints: Either of:

                - A :class:`.QueryResult`
                - List of :class:`.Datapoint` objects
                - List of dictionaries. Each dictionary should have fields ``id`` and ``download_url``.
                    ``id`` is the ID of the datapoint in the datasource.

            open_project: Automatically open the created Label Studio project in the browser.
            ignore_warning: Suppress the prompt-warning if you try to annotate too many datapoints at once.
            fields_to_embed: list of meta-data columns that will show up in Label Studio UI.
             if not specified all will be displayed.
            fields_to_exclude: list of meta-data columns that will not show up in Label Studio UI
        Returns:
            Link to open Label Studio in the browser
        """
        for f in (fields_to_embed or []) + (fields_to_exclude or []):
            if not self.has_field(f):
                raise FieldNotFoundError(f)

        if len(datapoints) == 0:
            logger.warning("No datapoints provided to be sent to annotation")
            return None
        elif len(datapoints) > dagshub.common.config.recommended_annotate_limit and not ignore_warning:
            force = prompt_user(
                f"You are attempting to annotate {len(datapoints)} datapoints at once - it's "
                f"recommended to only annotate up to "
                f"{dagshub.common.config.recommended_annotate_limit} "
                f"datapoints at a time."
            )
            if not force:
                return ""

        req_data = {
            "datasource_id": self.source.id,
            "datapoints": [],
            "ls_meta_excludes": fields_to_exclude,
            "ls_meta_includes": fields_to_embed,
        }

        for dp in datapoints:
            req_dict = {}
            if type(dp) is dict:
                req_dict["id"] = dp["datapoint_id"]
                req_dict["download_url"] = dp["download_url"]
            else:
                req_dict["id"] = dp.datapoint_id
                req_dict["download_url"] = dp.download_url
            req_data["datapoints"].append(req_dict)

        init_url = multi_urljoin(self.source.repoApi.data_engine_url, "annotations/init")
        resp = http_request("POST", init_url, json=req_data, auth=self.source.repoApi.auth)

        if resp.status_code != 200:
            logger.error(f"Error while sending request for annotation: {resp.content}")
            return None
        link = resp.json()["link"]

        # Do a raw print so it works in colab/jupyter
        print("Open the following link to start working on your annotation project:")
        print(link)

        if open_project:
            webbrowser.open_new_tab(link)
        return link

    def _launch_annotation_workspace(self):
        try:
            start_workspace_url = multi_urljoin(self.source.repoApi.annotations_url, "start")
            http_request("POST", start_workspace_url, auth=self.source.repoApi.auth)
        except:  # noqa
            pass

    def wait_until_ready(self, max_wait_time=300, fail_on_timeout=True):
        """
        Blocks until the datasource preprocessing is complete.

        Useful when you have just created a datasource and the initial scanning hasn't finished yet.

        Args:
            max_wait_time: Maximum time to wait in seconds
            fail_on_timeout: Whether to raise a RuntimeError or log a warning if the scan does not complete on time
        """

        # Start LS workspace to save time later in the flow
        self._launch_annotation_workspace()

        start = time.time()
        if max_wait_time:
            rich_console.log(f"Maximum waiting time set to {int(max_wait_time / 60)} minutes")
        spinner = rich_console.status("Waiting for datasource preprocessing to complete...")
        with spinner:
            while True:
                self.source.get_from_dagshub()
                if self.source.preprocessing_status == PreprocessingStatus.READY:
                    return

                if self.source.preprocessing_status == PreprocessingStatus.FAILED:
                    raise RuntimeError("Datasource preprocessing failed")

                if max_wait_time is not None and (time.time() - start) > max_wait_time:
                    if fail_on_timeout:
                        raise RuntimeError(
                            f"Time limit of {max_wait_time} seconds reached before processing was completed."
                        )
                    else:
                        logger.warning(
                            f"Time limit of {max_wait_time} seconds reached before processing was completed."
                        )
                        return

                time.sleep(1)

    def has_field(self, field_name: str) -> bool:
        """
        Checks if a metadata field ``field_name`` exists in the datasource.
        """
        reserved_searchable_fields = ["path"]
        fields = (f.name for f in self.fields)
        return field_name in reserved_searchable_fields or field_name in fields

    def __repr__(self):
        res = f"Datasource {self.source.name}"
        res += f"\n\tRepo: {self.source.repo}, path: {self.source.path}"
        if self.assigned_dataset:
            res += (
                f"\n\tAssigned Dataset: {self.assigned_dataset.dataset_name} "
                f"(Query was modified: {self.is_query_different_from_dataset})"
            )
        res += f"\n\t{self._query}"
        res += "\n\tFields:"
        for f in self.fields:
            res += f"\n\t\t{f}"
        return res + "\n"

    """ FUNCTIONS RELATED TO QUERYING
    These are functions that overload operators on the DataSet, so you can do pandas-like filtering
        ds = Dataset(...)
        queried_ds = ds[ds["value"] == 5]
    """

    def __getitem__(self, other: Union[slice, str, "Datasource", "Field"]):
        # Slicing - get items from the slice
        if type(other) is slice:
            return self.sample(other.start, other.stop)

        # Otherwise we're doing querying
        new_ds = self.__deepcopy__()
        if isinstance(other, (str, Field)):
            query_field: Field = Field(other) if type(other) is str else other
            other_query = QueryFilterTree(
                query_field.field_name,
                query_field.as_of_timestamp,
            )
            if not self.has_field(query_field.field_name):
                raise FieldNotFoundError(query_field.field_name)
            if self._query.filter.is_empty:
                new_ds._query.filter = other_query
            else:
                new_ds._query.compose("and", other_query)
            return new_ds
        # "index" is a datasource with a query - return the datasource inside
        # Example:
        #   ds = Dataset()
        #   filtered_ds = ds[ds["aaa"] > 5]
        #   filtered_ds2 = filtered_ds[filtered_ds["bbb"] < 4]
        # filtered_ds2 will be "aaa" > 5 AND "bbb" < 4
        if isinstance(other, Datasource):
            return other

    def __gt__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("gt", other)

    def __ge__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("ge", other)

    def __le__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("le", other)

    def __lt__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("lt", other)

    def __eq__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if other is None:
            return self.is_null()
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("eq", other)

    def __ne__(self, other: object):
        self._test_not_comparing_other_ds(other)
        if other is None:
            return self.is_not_null()
        if not isinstance(other, (int, float, str)):
            raise NotImplementedError
        return self.add_query_op("eq", other).add_query_op("not")

    def __invert__(self):
        return self.add_query_op("not")

    def __contains__(self, item):
        raise WrongOperatorError("Use `ds.contains(a)` for querying instead of `a in ds`")

    def contains(self, item: str):
        """
        Check if the filtering field contains the specified string item.

        :meta private:
        """
        if type(item) is not str:
            return WrongOperatorError(f"Cannot use contains with non-string value {item}")
        self._test_not_comparing_other_ds(item)
        return self.add_query_op("contains", item)

    def is_null(self):
        """
        Check if the filtering field is null.

        :meta private:
        """
        field = self._get_filtering_field()
        value_type = metadataTypeLookupReverse[field.valueType.value]
        return self.add_query_op("isnull", value_type())

    def is_not_null(self):
        """
        Check if the filtering field is not null.

        :meta private:
        """
        return self.is_null().add_query_op("not")

    def _get_filtering_field(self) -> MetadataFieldSchema:
        field_name = self.get_query().filter.column_filter
        if field_name is None:
            raise RuntimeError("The current query filter is not a field")
        for col in self.source.metadata_fields:
            if col.name == field_name:
                return col
        raise RuntimeError(f"Field {field_name} doesn't exist in the current uploaded metadata")

    def __and__(self, other: "Datasource"):
        return self.add_query_op("and", other)

    def __or__(self, other: "Datasource"):
        return self.add_query_op("or", other)

    # Prevent users from messing up their queries due to operator order
    # They always need to put the dataset query filters in parentheses, otherwise the binary and/or get executed before
    def __rand__(self, other):
        if type(other) is not Datasource:
            raise WrongOrderError(type(other))
        raise NotImplementedError

    def __ror__(self, other):
        if type(other) is not Datasource:
            raise WrongOrderError(type(other))
        raise NotImplementedError

    def add_query_op(
        self, op: str, other: Optional[Union[str, int, float, "Datasource", "QueryFilterTree"]] = None
    ) -> "Datasource":
        """
        Add a query operation to the current Datasource instance.

        Args:
            op (str): The operation to be performed in the query.
            other (Optional[Union[str, int, float, "Datasource", "DatasourceQuery"]], optional):
                The operand for the query operation. Defaults to None.

        Returns:
            Datasource: A new Datasource instance with the added query operation.

        :meta private:
        """
        new_ds = self.__deepcopy__()
        if type(other) is Datasource:
            other = other.get_query()
        new_ds._query.compose(op, other)
        return new_ds

    @staticmethod
    def _test_not_comparing_other_ds(other):
        if type(other) is Datasource:
            raise DatasetFieldComparisonError()


class MetadataContextManager:
    """
    Context manager for updating the metadata on a datasource.
    Batches the metadata changes, so they are being sent all at once.
    """

    def __init__(self, datasource: Datasource):
        self._datasource = datasource
        self._metadata_entries: List[DatapointMetadataUpdateEntry] = []
        self._multivalue_fields = datasource._get_multivalue_fields()

    def update_metadata(self, datapoints: Union[List[str], str], metadata: Dict[str, Any]):
        """
        Update metadata for the specified datapoints.

        .. note::
            If ``datapoints`` is a list, the same metadata is assigned to all the datapoints in the list.
            Call ``update_metadata()`` separately for each datapoint if you need to assign different metadata.

        Args:
            datapoints (Union[List[str], str]): A list of datapoints or a single datapoint path to update metadata for.
            metadata (Dict[str, Any]): A dictionary containing metadata key-value pairs to update.

        Example::

            with ds.metadata_context() as ctx:
                metadata = {
                    "episode": 5,
                    "has_baby_yoda": True,
                }

                # Attach metadata to a single specific file in the datasource.
                # The first argument is the filepath to attach metadata to, **relative to the root of the datasource**.
                ctx.update_metadata("images/005.jpg", metadata)

                # Attach metadata to several files at once:
                ctx.update_metadata(["images/006.jpg","images/007.jpg"], metadata)

        """
        if isinstance(datapoints, str):
            datapoints = [datapoints]

        field_value_types = {f.name: f.valueType for f in self._datasource.fields}

        for dp in datapoints:
            for k, v in metadata.items():
                if v is None:
                    continue
                if k in autogenerated_columns:
                    continue

                if type(v) is list:
                    if k not in self._multivalue_fields:
                        self._multivalue_fields.add(k)
                        # Promote all existing ones to multivalue
                        for e in self._metadata_entries:
                            if e.key == k:
                                e.allowMultiple = True
                    for sub_val in v:
                        value_type = field_value_types.get(k)
                        if value_type is None:
                            value_type = metadataTypeLookup[type(sub_val)]
                            field_value_types[k] = value_type
                        # Don't override bytes if they're not bytes - probably just undownloaded values
                        if value_type == MetadataFieldType.BLOB and type(sub_val) is not bytes:
                            continue

                        if type(v) is bytes:
                            sub_val = self.wrap_bytes(sub_val)
                        self._metadata_entries.append(
                            DatapointMetadataUpdateEntry(
                                url=dp,
                                key=k,
                                value=str(sub_val),
                                # todo: preliminary type check
                                valueType=value_type,
                                allowMultiple=k in self._multivalue_fields,
                            )
                        )

                else:
                    value_type = field_value_types.get(k)
                    if value_type is None:
                        value_type = metadataTypeLookup[type(v)]
                        field_value_types[k] = value_type
                    # Don't override bytes if they're not bytes - probably just undownloaded values
                    if value_type == MetadataFieldType.BLOB and type(v) is not bytes:
                        continue

                    if type(v) is bytes:
                        v = self.wrap_bytes(v)

                    self._metadata_entries.append(
                        DatapointMetadataUpdateEntry(
                            url=dp,
                            key=k,
                            value=str(v),
                            valueType=value_type,
                            # todo: preliminary type check
                            allowMultiple=k in self._multivalue_fields,
                        )
                    )

    @staticmethod
    def wrap_bytes(val: bytes) -> str:
        """
        Handles bytes values for uploading metadata
        The process is gzip -> base64

        :meta private:
        """
        compressed = gzip.compress(val)
        return base64.b64encode(compressed).decode("utf-8")

    def get_metadata_entries(self):
        return self._metadata_entries


@dataclass
class DatasourceQuery(DataClassJsonMixin):
    as_of: Optional[int] = field(default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL))
    select: Optional[List[Dict]] = field(default=None, metadata=config(exclude=exclude_if_none))
    filter: "QueryFilterTree" = field(
        default=QueryFilterTree(),
        metadata=config(field_name="query", encoder=QueryFilterTree.serialize, decoder=QueryFilterTree.deserialize),
    )
    order_by: Optional[List] = field(
        default=None, metadata=config(exclude=exclude_if_none, letter_case=LetterCase.CAMEL)
    )

    def __deepcopy__(self, memodict={}):
        other = DatasourceQuery(
            as_of=self.as_of,
            filter=self.filter.__deepcopy__(),
        )
        if self.select is not None:
            other.select = self.select.copy()
        if self.order_by is not None:
            other.order_by = self.order_by.copy()
        return other

    def compose(
        self, op: str, other: Optional[Union[str, int, float, "DatasourceQuery", "QueryFilterTree", "Datasource"]]
    ):
        other_query: Optional["DatasourceQuery"] = None
        # Extract the filter tree for composition
        if isinstance(other, (DatasourceQuery, Datasource)):
            if type(other) is Datasource:
                other = other.get_query()
            other_query = other
            other = other.filter

        self.filter.compose(op, other)

        # If composition went successfully, carry over the as_of, select and order_by from the other query
        if other_query is not None:
            if other_query.select is not None:
                self.select = other_query.select
            if other_query.as_of is not None:
                self.as_of = other_query.as_of
            if other_query.order_by is not None:
                self.order_by = other_query.order_by


@dataclass
class DatasourceSerializedState(DataClassJsonMixin):
    """
    Serialized state of the datasource.
    This should be enough to recreate the exact copy of the datasource back (with additional requests)

    Also carries additional information that might be useful for the user:
        - if the state of the datasource at the point of saving differed from the dataset
        - the timestamp of saving
        - link to open the datasource on DagsHub
    """

    repo: str
    """Repository this datasource is on"""
    datasource_id: Union[str, int]
    """ID of the datasource"""
    datasource_name: str
    """Name of the datasource"""
    query: DatasourceQuery
    """Query at the time of saving"""
    dataset_id: Optional[Union[str, int]] = None
    """ID of the assigned dataset"""
    dataset_name: Optional[str] = None
    """Name of the assigned dataset"""
    timestamp: Optional[float] = None
    """Timestamp of serialization"""
    modified: Optional[bool] = None
    """Does the query differ from the query in the assigned dataset"""
    link: Optional[str] = None
    """URL to open this datasource on DagsHub"""


@dataclass
class DatasetState:
    """
    Information about the Dataset.

    Dataset is a Datasource with a Query applied on it.
    """

    dataset_id: Union[str, int]
    """
    ID of the dataset
    """
    dataset_name: str
    """
    Name of the dataset
    """
    datasource_id: Union[str, int]
    """
    ID of the datasource with which this dataset is associated
    """
    query: DatasourceQuery
    """
    Query of this dataset
    """

    @staticmethod
    def from_dataset_query(
        dataset_id: Union[str, int], dataset_name: str, datasource_id: Union[str, int], dataset_query: Union[Dict, str]
    ) -> "DatasetState":
        if type(dataset_query) is str:
            dataset_query = json.loads(dataset_query)
        query = DatasourceQuery.from_dict(dataset_query)
        res = DatasetState(dataset_id=dataset_id, dataset_name=dataset_name, datasource_id=datasource_id, query=query)
        return res

    @staticmethod
    def from_gql_dataset_result(dataset_result: "DatasetResult") -> "DatasetState":
        return DatasetState.from_dataset_query(
            dataset_id=dataset_result.id,
            dataset_name=dataset_result.name,
            datasource_id=dataset_result.datasource.id,
            dataset_query=dataset_result.datasetQuery,
        )



from typing import List, Optional, Union

from dagshub.common.analytics import send_analytics_event
from dagshub.data_engine.client.data_client import DataClient
from dagshub.data_engine.client.models import DatasetResult
from dagshub.data_engine import datasources
from dagshub.data_engine.model.datasource import Datasource, DEFAULT_MLFLOW_ARTIFACT_NAME, DatasetState
from dagshub.data_engine.model.datasource_state import DatasourceState
from dagshub.data_engine.model.errors import DatasetNotFoundError


def get_datasets(repo: str) -> List[Datasource]:
    """
    Get all datasources that exist on the repo

    Args:
        repo: Repo in ``<owner>/<reponame>`` format

    Returns:
        list(Datasource): All datasets of the repo
    """
    return _get_datasets(repo)


def get_dataset(repo: str, name: Optional[str] = None, id: Optional[Union[int, str]] = None) -> Datasource:
    """
    Get specific dataset by name or id

    Args:
        repo: Repo in ``<owner>/<reponame>`` format
        name: Name of the dataset
        id: ID of the dataset

    Returns:
        Datasource: Found dataset

    Raises:
        DatasetNotFoundError: No dataset found with this name or id
    """
    assert name is not None or id is not None

    sources = _get_datasets(repo, name, id)
    if len(sources) == 0:
        raise DatasetNotFoundError(repo, id, name)
    return sources[-1]


def get_dataset_from_file(path: str) -> Datasource:
    """
    Load a dataset from a local file.

    This is a copy of
    :func:`datasources.get_datasource_from_file()<dagshub.data_engine.datasources.get_datasource_from_file>`

    Args:
        path: Path to the ``.dagshub`` file with the relevant dataset

    Returns:
        dataset that was logged to the file
    """
    return datasources.get_datasource_from_file(path)


def get_from_mlflow(run=None, artifact_name=DEFAULT_MLFLOW_ARTIFACT_NAME) -> Datasource:
    """
    Load a dataset from an MLflow run.

    To save a datasource to MLflow, use
    :func:`Datasource.log_to_mlflow()<dagshub.data_engine.model.datasource.Datasource.log_to_mlflow>`.

    This is a copy of :func:`datasources.get_from_mlflow()<dagshub.data_engine.datasources.get_from_mlflow>`

    Args:
        run: Run or ID of the MLflow run to load the datasource from.
            If ``None``, gets it from the current active run.
        artifact_name: Name of the artifact in the run.
    """
    return datasources.get_from_mlflow(run, artifact_name)


def _get_datasets(repo: str, name: Optional[str] = None, id: Optional[Union[int, str]] = None) -> List[Datasource]:
    send_analytics_event("Client_DataEngine_getDatasets")
    client = DataClient(repo)
    sources = client.get_datasets(id, name)
    return [_from_gql_result(repo, source) for source in sources]


def _from_gql_result(repo: str, dataset_result: "DatasetResult") -> "Datasource":
    ds = Datasource(DatasourceState.from_gql_result(repo, dataset_result.datasource))

    dataset_state = DatasetState.from_gql_dataset_result(dataset_result)
    ds.load_from_dataset_state(dataset_state)
    return ds



class WrongOrderError(Exception):
    def __init__(self, other_type: type):
        super().__init__()
        self.other_type = other_type

    def __str__(self):
        return (
            f"Can't have a dataset to the right of {self.other_type}.\r\n"
            f"Make sure to use parentheses to chain logical and/or operators.\r\n"
            f"Example: `ds[(ds['col1'] > 1) & (ds['col2'] < 3)])`"
        )


class DatasetFieldComparisonError(Exception):
    def __str__(self):
        return (
            "Can't compare two fields in a dataset between each other.\r\n"
            "Querying only supports comparisons with primitives (int/str/float)"
        )


class WrongOperatorError(Exception):
    pass


class FieldNotFoundError(Exception):
    def __init__(self, field: str):
        super().__init__()
        self.field = field

    def __str__(self):
        return f"Field {self.field} does not exist on this datasource"


class DataEngineGqlError(Exception):
    def __init__(self, original_exception, support_id):
        super().__init__()
        self.original_exception = original_exception
        self.support_id = support_id

    def __str__(self):
        return (
            f"Original exception: {self.original_exception.__class__.__name__} - {self.original_exception} \n"
            f"Support Id: {self.support_id}"
        )


class DatasourceAlreadyExistsError(Exception):
    def __init__(self, datasource):
        super().__init__()
        self.datasource = datasource

    def __str__(self):
        return f"Datasource with name {self.datasource.name} already exists in repository {self.datasource.repo}"


class DatasourceNotFoundError(Exception):
    def __init__(self, datasource):
        super().__init__()
        self.datasource = datasource

    def __str__(self):
        return (
            f"Datasource with name {self.datasource.name} or id {self.datasource.id} not found "
            f"in repository {self.datasource.repo}"
        )


class DatasetNotFoundError(Exception):
    def __init__(self, repo, id, name):
        super().__init__()
        self.repo = repo
        self.id = id
        self.name = name

    def __str__(self):
        return f"Dataset with name {self.name} or id {self.id} not found " f"in repository {self.repo}"



import sys
import setuptools
import os.path


# Thank you pip contributors
def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

install_requires = [
    "PyYAML>=5",
    "fusepy>=3",
    "appdirs>=1.4.4",
    "click>=8.0.4",
    "httpx~=0.23.0",
    "GitPython>=3.1.29",
    "rich~=13.1.0",
    # Need to keep dacite version in lockstep with voxel, otherwise stuff breaks on their end
    "dacite~=1.6.0",
    "tenacity~=8.2.2",
    "gql[requests]",
    "dataclasses-json",
    "pandas",
    "treelib~=1.6.4",
    "pathvalidate~=3.0.0",
    "python-dateutil",
    "tenacity~=8.2.3",
    "boto3",
]

extras_require = {
    "jupyter": ["rich[jupyter]~=13.1.0"],
    "fuse": ["fusepy>=3"],
}

# Polyfills for Python 3.7
if sys.version_info.major == 3 and sys.version_info.minor == 7:
    install_requires += [
        "cached-property==1.5.2",
        "typing_extensions",
    ]

packages = setuptools.find_packages(exclude=["tests", "tests.*"])

setuptools.setup(
    name="dagshub",
    version=get_version("dagshub/__init__.py"),
    author="DagsHub",
    author_email="contact@dagshub.com",
    description="DagsHub client libraries",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DagsHub/client",
    packages=packages,
    install_requires=install_requires,
    extras_require=extras_require,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    entry_points={"console_scripts": ["dagshub = dagshub.common.cli:cli"]},
)



import urllib.parse
from pathlib import Path
from typing import Optional, Union, Tuple

from dagshub.common import config
from dagshub.common.api import RepoAPI
from dagshub.common.errors import DagsHubRepoNotFoundError
from dagshub.common.util import lazy_load

git = lazy_load("git")


def parse_dagshub_remote(remote_url: urllib.parse.ParseResult, host_url: urllib.parse.ParseResult) -> Optional[str]:
    """
    Tries to parse the remote url, extracting the name of a DagsHub repo from it and returning it

    Args:
        remote_url: parsed URL of the remote
        host_url: parsed URL of the DagsHub host

    Returns:
        Repository name in the `<user>/<repo>` format
    """
    if remote_url.hostname != host_url.hostname:
        return None

    # Check for the host prefix and the path ending with .git
    if not (remote_url.path.startswith(host_url.path) and remote_url.path.endswith(".git")):
        return None

    subpath = remote_url.path[len(host_url.path) :]
    # Should leave the subpath of the host, if the remote is correct it should be just "/user/repo.git"
    subpath = subpath.lstrip("/")
    if subpath.count("/") > 1:
        return None
    return subpath[: -len(".git")]


def determine_repo(path: Optional[Union[str, Path]] = None, host: Optional[str] = None) -> Tuple[RepoAPI, str]:
    """
    Tries to find a DagsHub repository in the specified path by traversing up the tree
    and looking for a git repo with the DagsHub remote.

    Args:
        path: Path where to look for the repo. If None, looks in current working directory
        host: DagsHub host, defaults to https://dagshub.com

    Returns:
        RepoAPI object of the repo + name of the current branch
    """
    if path is None:
        path = Path.cwd()
    else:
        path = Path(path)

    if host is None:
        host = config.host

    try:
        repo = git.Repo(path, search_parent_directories=True)
    except git.InvalidGitRepositoryError as ex:
        raise DagsHubRepoNotFoundError(path) from ex

    repo_name: Optional[str] = None
    host_url = urllib.parse.urlparse(host)
    for remote in repo.remotes:
        repo_name = parse_dagshub_remote(urllib.parse.urlparse(remote.url), host_url)
        if repo_name is not None:
            break

    if repo_name is None:
        raise DagsHubRepoNotFoundError(path)

    return RepoAPI(repo_name, host=host), repo.active_branch.name
